@@ -1,71 +1,28 @@
+# -*- coding: utf-8 -*-
 # This file contains code dealing specifically with converting Mercurial
 # repositories to Git repositories. Code in this file is meant to be a generic
 # library and should be usable outside the context of hg-git or an hg command.
 
+import io
 import os
 import stat
 
+import dulwich.config as dulcfg
 import dulwich.objects as dulobjs
 from mercurial import (
-    util as hgutil,
+    subrepoutil,
+    encoding,
+    error,
 )
 
-import compat
-import util
 
-def parse_subrepos(ctx):
-    sub = util.OrderedDict()
-    if '.hgsub' in ctx:
-        sub = util.parse_hgsub(ctx['.hgsub'].data().splitlines())
-    substate = util.OrderedDict()
-    if '.hgsubstate' in ctx:
-        substate = util.parse_hgsubstate(
-            ctx['.hgsubstate'].data().splitlines())
-    return sub, substate
-
-
-def audit_git_path(ui, path):
-    r"""Check for path components that case-fold to .git.
-
-    >>> class fakeui(object):
-    ...     def configbool(*args):
-    ...         return False
-    ...     def warn(self, s):
-    ...         print s
-    >>> u = fakeui()
-    >>> audit_git_path(u, 'foo/git~100/wat')
-    warning: path 'foo/git~100/wat' contains a potentially dangerous path component.
-    It may not be legal to check out in Git.
-    It may also be rejected by some git server configurations.
-    <BLANKLINE>
-    >>> audit_git_path(u, u'foo/.gi\u200ct'.encode('utf-8'))
-    warning: path 'foo/.gi\xe2\x80\x8ct' contains a potentially dangerous path component.
-    It may not be legal to check out in Git.
-    It may also be rejected by some git server configurations.
-    <BLANKLINE>
-    >>> audit_git_path(u, 'this/is/safe')
-    """
-    dangerous = False
-    for c in path.split(os.path.sep):
-        if compat.hfsignoreclean(c) == '.git':
-            dangerous = True
-            break
-        elif '~' in c:
-            base, tail = c.split('~', 1)
-            if tail.isdigit() and base.upper().startswith('GIT'):
-                dangerous = True
-                break
-    if dangerous:
-        if compat.config(ui, 'bool', 'git', 'blockdotgit'):
-            raise hgutil.Abort(
-                ('Refusing to export likely-dangerous path %r' % path),
-                hint=("If you need to continue, read about CVE-2014-9390 and "
-                      "then set '[git] blockdotgit = false' in your hgrc."))
-        ui.warn('warning: path %r contains a potentially dangerous path '
-                'component.\n'
-                'It may not be legal to check out in Git.\n'
-                'It may also be rejected by some git server configurations.\n'
-                % path)
+def flags2mode(flags):
+    if b'l' in flags:
+        return 0o120000
+    elif b'x' in flags:
+        return 0o100755
+    else:
+        return 0o100644
 
 
 class IncrementalChangesetExporter(object):
@@ -134,16 +91,16 @@ class IncrementalChangesetExporter(object):
             return
         dirkind = stat.S_IFDIR
         # depth-first order, chosen arbitrarily
-        todo = [('', store[commit.tree])]
+        todo = [(b'', store[commit.tree])]
         while todo:
             path, tree = todo.pop()
             self._dirs[path] = tree
-            for entry in tree.iteritems():
+            for entry in tree.items():
                 if entry.mode == dirkind:
-                    if path == '':
+                    if path == b'':
                         newpath = entry.path
                     else:
-                        newpath = path + '/' + entry.path
+                        newpath = path + b'/' + entry.path
                     todo.append((newpath, store[entry.sha]))
 
     @property
@@ -152,7 +109,76 @@ class IncrementalChangesetExporter(object):
 
         This is needed to construct a Git commit object.
         """
-        return self._dirs[''].id
+        return self._dirs[b''].id
+
+    def audit_path(self, path):
+        r"""Check for path components that case-fold to .git.
+
+        Returns ``True`` for normal paths. The results for insecure
+        paths depend on the ``hggit.invalidpaths`` configuration in
+        ``ui``:
+
+        ``abort`` means insecure paths abort the conversion —
+        this was the default prior to 1.0.
+
+        ``keep`` means issue a warning and keep the path,
+        returning ``True``.
+
+        Anything else, but documented as ``skip``, means issue a
+        warning and disregard the path, returning ``False``.
+
+        """
+        ui = self._hg.ui
+
+        dangerous = False
+
+        for c in path.split(b'/'):
+            if encoding.hfsignoreclean(c) == b'.git':
+                dangerous = True
+                break
+            elif b'~' in c:
+                base, tail = c.split(b'~', 1)
+                if tail.isdigit() and base.upper().startswith(b'GIT'):
+                    dangerous = True
+                    break
+        if dangerous:
+            opt = ui.config(b'hggit', b'invalidpaths')
+
+            # escape the path when printing it out
+            prettypath = path.decode('latin1').encode('unicode-escape')
+
+            if opt == b'abort':
+                raise error.Abort(
+                    b"invalid path '%s' rejected by configuration" % prettypath,
+                    hint=b"see 'hg help config.hggit.invalidpaths for details",
+                )
+            elif opt == b'keep':
+                ui.warn(
+                    b"warning: path '%s' contains an invalid path component\n"
+                    % prettypath,
+                )
+                return True
+            else:
+                # undocumented: just let anything else mean "skip"
+                ui.warn(b"warning: skipping invalid path '%s'\n" % prettypath)
+                return False
+        elif path == b'.gitmodules':
+            ui.warn(
+                b"warning: ignoring modifications to '%s' file; "
+                b"please use '.hgsub' instead\n" % path
+            )
+            return False
+        else:
+            return True
+
+    def filter_unsafe_paths(self, paths):
+        """Return a copy of the given list, with dangerous paths removed.
+
+        Please note that the configuration can suppress the removal;
+        see above.
+
+        """
+        return [path for path in paths if self.audit_path(path)]
 
     def update_changeset(self, newctx):
         """Set the tree to track a new Mercurial changeset.
@@ -188,17 +214,22 @@ class IncrementalChangesetExporter(object):
         # The only reliable way to get the full set of changes is by looking at
         # the full manifest. And, the easy way to compare two manifests is
         # localrepo.status().
-        modified, added, removed = self._hg.status(self._ctx, newctx)[0:3]
+        status = self._hg.status(self._ctx, newctx)
+        modified, added, removed = map(
+            self.filter_unsafe_paths,
+            (status.modified, status.added, status.removed),
+        )
 
         # We track which directories/trees have modified in this update and we
         # only export those.
         dirty_trees = set()
 
+        gitmodules = b''
         subadded, subremoved = [], []
 
         for s in modified, added, removed:
-            if '.hgsub' in s or '.hgsubstate' in s:
-                subadded, subremoved = self._handle_subrepos(newctx)
+            if b'.hgsub' in s or b'.hgsubstate' in s:
+                gitmodules, subadded, subremoved = self._handle_subrepos(newctx)
                 break
 
         # We first process subrepo and file removals so we can prune dead
@@ -207,12 +238,16 @@ class IncrementalChangesetExporter(object):
             self._remove_path(path, dirty_trees)
 
         for path in removed:
-            if path == '.hgsubstate' or path == '.hgsub':
+            if path == b'.hgsubstate' or path == b'.hgsub':
                 continue
 
             self._remove_path(path, dirty_trees)
 
         for path, sha in subadded:
+            if b'.gitmodules' in newctx:
+                modified.append(b'.gitmodules')
+            else:
+                added.append(b'.gitmodules')
             d = os.path.dirname(path)
             tree = self._dirs.setdefault(d, dulobjs.Tree())
             dirty_trees.add(d)
@@ -222,20 +257,27 @@ class IncrementalChangesetExporter(object):
         # corresponding Git blob and its tree entry. We emit the blob
         # immediately and update trees to be aware of its presence.
         for path in set(modified) | set(added):
-            audit_git_path(self._hg.ui, path)
-            if path == '.hgsubstate' or path == '.hgsub':
+            if path == b'.hgsubstate' or path == b'.hgsub':
                 continue
 
             d = os.path.dirname(path)
             tree = self._dirs.setdefault(d, dulobjs.Tree())
             dirty_trees.add(d)
 
-            fctx = newctx[path]
+            if path == b'.gitmodules':
+                blob = dulobjs.Blob.from_string(gitmodules)
+                entry = dulobjs.TreeEntry(
+                    path,
+                    flags2mode(newctx[b'.hgsub'].flags()),
+                    blob.id,
+                )
+            else:
+                fctx = newctx[path]
+
+                entry, blob = self.tree_entry(fctx)
 
-            func = IncrementalChangesetExporter.tree_entry
-            entry, blob = func(fctx, self._blob_cache)
             if blob is not None:
-                yield (blob, fctx.filenode())
+                yield blob
 
             tree.add(*entry)
 
@@ -245,7 +287,7 @@ class IncrementalChangesetExporter(object):
         # dulwich.index.commit_tree(), which builds new Tree instances for each
         # series of blobs.
         for obj in self._populate_tree_entries(dirty_trees):
-            yield (obj, None)
+            yield obj
 
         self._ctx = newctx
 
@@ -279,7 +321,7 @@ class IncrementalChangesetExporter(object):
             return
 
         # Now we traverse up to the parent and delete any references.
-        if path == '':
+        if path == b'':
             return
 
         basename = os.path.basename(path)
@@ -302,20 +344,20 @@ class IncrementalChangesetExporter(object):
             # The parent tree is empty. Se, we can delete it.
             del self._dirs[parent]
 
-            if parent == '':
+            if parent == b'':
                 return
 
             basename = os.path.basename(parent)
             parent = os.path.dirname(parent)
 
     def _populate_tree_entries(self, dirty_trees):
-        self._dirs.setdefault('', dulobjs.Tree())
+        self._dirs.setdefault(b'', dulobjs.Tree())
 
         # Fill in missing directories.
-        for path in self._dirs.keys():
+        for path in list(self._dirs):
             parent = os.path.dirname(path)
 
-            while parent != '':
+            while parent != b'':
                 parent_tree = self._dirs.get(parent, None)
 
                 if parent_tree is not None:
@@ -327,7 +369,7 @@ class IncrementalChangesetExporter(object):
         for dirty in list(dirty_trees):
             parent = os.path.dirname(dirty)
 
-            while parent != '':
+            while parent != b'':
                 if parent in dirty_trees:
                     break
 
@@ -335,7 +377,7 @@ class IncrementalChangesetExporter(object):
                 parent = os.path.dirname(parent)
 
         # The root tree is always dirty but doesn't always get updated.
-        dirty_trees.add('')
+        dirty_trees.add(b'')
 
         # We only need to recalculate and export dirty trees.
         for d in sorted(dirty_trees, key=len, reverse=True):
@@ -347,7 +389,7 @@ class IncrementalChangesetExporter(object):
 
             yield tree
 
-            if d == '':
+            if d == b'':
                 continue
 
             parent_tree = self._dirs[os.path.dirname(d)]
@@ -362,8 +404,8 @@ class IncrementalChangesetExporter(object):
             parent_tree[os.path.basename(d)] = (stat.S_IFDIR, tree.id)
 
     def _handle_subrepos(self, newctx):
-        sub, substate = parse_subrepos(self._ctx)
-        newsub, newsubstate = parse_subrepos(newctx)
+        state = subrepoutil.state(self._ctx, self._hg.ui)
+        newstate = subrepoutil.state(newctx, self._hg.ui)
 
         # For each path, the logic is described by the following table. 'no'
         # stands for 'the subrepo doesn't exist', 'git' stands for 'git
@@ -382,20 +424,28 @@ class IncrementalChangesetExporter(object):
         # 'added' is both modified and added
         added, removed = [], []
 
-        def isgit(sub, path):
-            return path not in sub or sub[path].startswith('[git]')
+        if b'.gitmodules' not in newctx:
+            config = dulcfg.ConfigFile()
+        else:
+            with io.BytesIO(newctx[b'.gitmodules'].data()) as buf:
+                config = dulcfg.ConfigFile.from_file(buf)
 
-        for path, sha in substate.iteritems():
-            if not isgit(sub, path):
+        submodules = {
+            path: name for (path, url, name) in dulcfg.parse_submodules(config)
+        }
+
+        for path, (remote, sha, t) in state.items():
+            if t != b'git':
                 # old = hg -- will be handled in next loop
                 continue
             # old = git
-            if path not in newsubstate or not isgit(newsub, path):
+            if path not in newstate or newstate[path][2] != b'git':
                 # new = hg or no, case (2) or (3)
                 removed.append(path)
+                submodules.pop(path, None)
 
-        for path, sha in newsubstate.iteritems():
-            if not isgit(newsub, path):
+        for path, (remote, sha, t) in newstate.items():
+            if t != b'git':
                 # new = hg or no; the only cases we care about are handled
                 # above
                 continue
@@ -403,32 +453,36 @@ class IncrementalChangesetExporter(object):
             # case (1)
             added.append((path, sha))
 
-        return added, removed
+            section = b"submodule", submodules.get(path, path)
+            config.set(section, b"path", path)
+            config.set(section, b"url", remote)
 
-    @staticmethod
-    def tree_entry(fctx, blob_cache):
+        with io.BytesIO() as buf:
+            config.write_to_file(buf)
+            gitmodules = buf.getvalue()
+
+        return gitmodules, added, removed
+
+    def tree_entry(self, fctx):
         """Compute a dulwich TreeEntry from a filectx.
 
-        A side effect is the TreeEntry is stored in the passed cache.
+        A side effect is the TreeEntry is stored in the blob cache.
 
         Returns a 2-tuple of (dulwich.objects.TreeEntry, dulwich.objects.Blob).
         """
-        blob_id = blob_cache.get(fctx.filenode(), None)
+        blob_id = self._blob_cache.get(fctx.filenode(), None)
         blob = None
 
         if blob_id is None:
             blob = dulobjs.Blob.from_string(fctx.data())
             blob_id = blob.id
-            blob_cache[fctx.filenode()] = blob_id
+            self._blob_cache[fctx.filenode()] = blob_id
 
-        flags = fctx.flags()
-
-        if 'l' in flags:
-            mode = 0120000
-        elif 'x' in flags:
-            mode = 0100755
-        else:
-            mode = 0100644
-
-        return (dulobjs.TreeEntry(os.path.basename(fctx.path()), mode,
-                                  blob_id), blob)
+        return (
+            dulobjs.TreeEntry(
+                os.path.basename(fctx.path()),
+                flags2mode(fctx.flags()),
+                blob_id,
+            ),
+            blob,
+        )
