@@ -422,22 +422,27 @@ def has_dulwich():
     except ImportError:
         return False
 
-@checkvers(
-    "dulwich", "Dulwich >= %s", [
-        '%d.%d.%d' % vers
-        for vers in (
-            (0, 20, 37),
-            (0, 20, 44),
-            (0, 21, 0),
-        )
-    ]
-)
+
+@checkvers("dulwich", "Dulwich >= %s", ['%d.%d.%d' % vers for vers in ()])
 def has_dulwich_range(v):
     import dulwich
 
     return dulwich.__version__ >= tuple(map(int, v.split('.')))
 
 
+@check("dulwich-rust", "Dulwich Python library using Rust bindings")
+def has_dulwich_rust():
+    try:
+        from dulwich import objects, __version__
+
+        return (
+            __version__ > (0, 22, 0)
+            and objects._parse_tree_py is not objects.parse_tree
+        )
+    except ImportError:
+        return False
+
+
 @check("pygit2", "pygit2 Python library")
 def has_pygit2():
     try:
@@ -1146,6 +1151,7 @@ def has_emacs():
 def has_black():
     try:
         import black
+
         version = black.__version__
     except ImportError:
         version = None
@@ -1200,6 +1206,8 @@ def has_unicode_filesystem():
             return False
     except Exception:
         return True
+
+
 @check("bigendian", "big-endian CPU")
 def has_bigendian():
     return sys.byteorder == 'big'
